// ================================================================
// Copyright (c) Meta Platforms, Inc. and affiliates.
// ================================================================

#ifndef TORBEN_H
#define TORBEN_H

namespace facebook {
namespace pdq {
namespace hashing {

/**
 * The following code is public domain.
 * Algorithm by Torben Mogensen, implementation by N. Devillard.
 */
float torben(float m[], int n);
} // namespace hashing
} // namespace pdq
} // namespace facebook

#endif // TORBEN_H
