// ================================================================
// Copyright (c) Meta Platforms, Inc. and affiliates.
// ================================================================

#ifndef PDQBASETYPES_H
#define PDQBASETYPES_H
#include <cstdint>
namespace facebook {
namespace pdq {
namespace hashing {

using Hash8 = std::uint8_t;
using Hash16 = std::uint16_t;
using Hash64 = std::uint64_t;

} // namespace hashing
} // namespace pdq
} // namespace facebook

#endif // PDQBASETYPES_H
